#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "main.h"

typedef enum {FALSE=0, TRUE=1} boolean;

static boolean lotsaspace = FALSE;
static long lots = 10000;
static boolean dham = FALSE;
static boolean idebug = FALSE;
#ifdef UNBOXED
static boolean unboxed = TRUE;
#else
static boolean unboxed = FALSE;
#endif
static boolean verbose = FALSE;
static long long superstepSleepTime = 0L;
static char *filename = NULL;
static char *arch = NULL;

static void run(char *cmd) {
	fprintf(stdout, "-------\nlaunching: %s\n", cmd);
	int err = system(cmd);
	if(err != 0) {
		fprintf(stderr, "Command failed (%d)\n", err);
		exit(1);
	}
}

static void showHelp(const char *applicationName) {
	fprintf(stdout,"%s: [-lotsaspace | --help ] [<infile>]\n", applicationName);
	fprintf(stdout,"\t -lotsaspace   Compile with lots of space (heap=%ld, stack=%ld, wire=%ld).\n",lots,lots,lots);
	fprintf(stdout,"\t -lots <long>  Amount of extra space to add for -lotsaspace (currently: %ld)",lots);
	fprintf(stdout,"\t -dham         Dump final internal ham format (as used by ham2c), then stop.\n");
	fprintf(stdout,"\t -g            Generate code with interactive debug support.\n");
	fprintf(stdout,"\t -nub          Generate non-un-boxed code.\n");
	fprintf(stdout,"\t -box          Generate boxed code.\n");
	fprintf(stdout,"\t -N <long>     Add nanosleep(<long>) to superstep.\n");
	fprintf(stdout,"\t -v            Verbose.\n");
	fprintf(stdout,"\t -arch <arch>  Specify the architecture, m32 or nix[default].\n");
	fprintf(stdout,"\t --help        Show this information.\n");
	fprintf(stdout,"\t <infile>      Specify the Hume/Ham/C file to compile.\n");
}

static void processArgs(int argc, const char * argv[]) {
	int i;
	for(i = 1; i < argc; i++) {
		if(argv[i][0] == '-') {
			if((strncmp(argv[i], "-h", 2) == 0) || (strncmp(argv[i], "--h", 3) == 0)) {
				showHelp(argv[0]);
				exit(0);
			} else if(strcmp(argv[i], "-lotsaspace") == 0) {
				lotsaspace = TRUE;
			} else if(strcmp(argv[i], "-dham") == 0) {
				dham = TRUE;
			} else if(strcmp(argv[i], "-g") == 0) {
				idebug = TRUE;
			} else if(strcmp(argv[i], "-v") == 0) {
				verbose = TRUE;
			} else if(strcmp(argv[i], "-nub") == 0) {
				unboxed = FALSE;
			} else if(strcmp(argv[i], "-box") == 0) {
				unboxed = TRUE;
			} else if((strcmp(argv[i], "-N") == 0) && (i+1 < argc)) {
				superstepSleepTime = atoll(argv[i+1]);
				i++;
			} else if((strcmp(argv[i], "-lots") == 0) && (i+1 < argc)) {
				lots = atoll(argv[i+1]);
				i++;
			} else if((strcmp(argv[i], "-arch") == 0) && (i+1 < argc)) {
				arch = (char*)argv[i+1];
				i++;
			} else {
				fprintf(stderr, "Invalid option %s\n", argv[i]);
				exit(1);
			}
		} else if(filename) {
			fprintf(stderr, "Please specify only one file\n");
			exit(1);
		} else
			filename = (char*)argv[i];
	}
	if(!filename) {
		fprintf(stderr, "Please specify a file to compile\n");
		exit(1);
	}
}


#define MAX_BUFF 2000

static void cleanStubs() {
	const char *stubsName = "humestubs.c";
	const char *oldStubsName = "humestubs.c.bak";
	FILE *inFile, *outFile;
	char buff[MAX_BUFF];
	int line = 0;
	if(rename(stubsName, oldStubsName) != 0) {
	  	fprintf(stderr, "Unable to use '%s'\n", stubsName);
		exit(1);
	}
	inFile = fopen(oldStubsName, "r");
	outFile = fopen(stubsName, "w");
	if(!outFile) {
		fprintf(stderr, "Unable to create new '%s'\n", stubsName);
		exit(1);
	} 
	while(fgets(buff, MAX_BUFF, inFile) != NULL) {
		int len = strlen(buff) - 1;
		if(buff[len] == '\n') {
			line++;
		} else {
			fprintf(stderr, "%d: Input line too long to parse\n", line);
			exit(1);
		}
		if(strncmp(buff, "#include \"am.h\" /*", 18) == 0) {
			strcpy(buff, "#include \"ccall.h\"\n"); /* correct line */
		} else if(strncmp(buff, "unsigned nccalls =", 18) == 0) {
			continue; /* remove line */
		} else if(strncmp(buff, "CCALL ccalls[] = ", 17) == 0) {
			continue; /* remove line */
		} else if(strncmp(buff, "/*", 2) == 0) {
			continue; /* skip comments */
		}
		fputs(buff, outFile);
	}
	fputs("/* This file has been cleaned up by humec */\n", outFile);
	fclose(outFile);
	fclose(inFile);
	unlink(oldStubsName); /* tidy up */
}

int main(int argc, const char* argv[]) {	
	char *type;
	char buff[4000]; /* bad, but it will do */
	processArgs(argc, argv);
	
	type = strrchr(filename, '.'); 
	if(!type) {
		type = ".hume";
	} else {
		char *temp = (char*)malloc(1+type-filename);
		strncpy(temp, filename, type-filename);
		temp[type-filename] = '\0';
		filename = temp;
	}

	if(strcmp(type, ".hume") == 0) {
		char *phamc = getenv("PHAMC"); 
		if(!phamc) phamc = PHAMC;
		sprintf(buff, "%s -t -v %s.hume", phamc, filename);
		run(buff);
		type = ".ham";
		
		/* phamc chucks things into the current working directory */
		{	char *p = strrchr(filename, '/');
			if(p) filename = p+1;
		}
		cleanStubs();
	}

	if(strcmp(type, ".ham") == 0) {
		char *ham2c = getenv("HAM2C"); 
		if(!ham2c) ham2c = HAM2C;
		sprintf(buff, "%s %s.ham", ham2c, filename);
		if(dham) strcat(buff, " -dham");
		if(lotsaspace) sprintf(buff+strlen(buff)," -mh %ld -mw %ld -ms %ld",lots,lots,lots);
		if(idebug) strcat(buff, " -g");
		if(!unboxed) strcat(buff, " -nub");
		if(verbose) strcat(buff, " -v");
		if(superstepSleepTime) sprintf(buff+strlen(buff)," -N %lld",superstepSleepTime);
		if(arch) sprintf(buff+strlen(buff), " -arch %s", arch);
		run(buff);
		if(dham) exit(0);
		type = ".c";
	}
	
	if(strcmp(type, ".c") == 0) {
		FILE *file = fopen("humestubs.c", "r");
		if(file) fclose(file);

		sprintf(buff, "%s %s.c", CC, filename);
		if(file) sprintf(buff+strlen(buff), " humestubs.c");
		if(unboxed) strcat(buff, " -DUNBOXED");
		sprintf(buff+strlen(buff), " -I %s %s/rts.ar -o %s", RTSDIR, RTSDIR, filename);
		run(buff);
		type = "";
	}

	if(strcmp(type, "") != 0) {
		fprintf(stderr, "unknown file type\n");
		exit(1);
	}
	return 0;
}
