#ifndef _WITHIN_H_
#define _WITHIN_H_
#include "am.h"
#include "when.h"

#ifdef NDEBUG
#define initWithin()
#define closeWithin()
#define clearWithin()
#define pushWithin(u,exception)
#define popWithin()
#define pushWithinStack(limit,exception)
#define popWithinStack()
#define pushWithinHeap(limit,exception)
#define popWithinHeap()
#else
/*
 * Within handling of time, heap, and space...
 *
 */

/* Only need to init once */
void initWithin();
void closeWithin();

/* Will need to call this in the exception handler to ensure withins don't interfer */
void clearWithin();

/* Providing these are balanced then there is no need for a clearWithin() */
void pushWithin(usecs u, uint16 exception);
void popWithin();

void pushWithinStack(uint32 limit,  uint16 exception);
void popWithinStack();

void pushWithinHeap(uint32 limit,  uint16 exception);
void popWithinHeap();

#endif /*NDEBUG*/

extern uint16 currentStackException;
extern uint16 currentHeapException;



/*
 * Provide support for the suspension of time
 * both in terms of interrupts and in terms of instrumentation
 * - this is only necessary to suppport debugging
 *
 */

#ifdef SUPPORT_DEBUG
/* this structure should be kept opaque */
struct freeze_t;
typedef struct freeze_t freeze_t;
struct freeze_t {
	when_t start;
	struct itimerval timer;
};

void freezeTime(freeze_t *frozen);
void unfreezeTime(freeze_t *frozen);
#endif /*SUPPORT_DEBUG*/

#endif /* _WITHIN_H_ */
